#ifndef _DHT11_INTERNAL_H_
#define _DHT11_INTERNAL_H_

//sunxi_gpio
#define SUNXI_GPIO_BASE       (0x01c20800)
#define SUNXI_GPIO_LM_BASE    (0x01f02c00)
#define MAP_SIZE	          (4096*2)
#define MAP_MASK	          (MAP_SIZE - 1)


#define IOMEM(x) ((void __force __iomem *)(x))
#ifndef IO_ADDRESS
#define IO_ADDRESS(x) ((x) + 0xf0000000)
#endif
#ifndef __io_address
#define __io_address(n) IOMEM(IO_ADDRESS(n))
#endif

// IO memory map base address, discovered by searching through Armbian sources
#define SUNXI_IOMEM(n) (0xf0000000 + (n))
#define PIO_BASE SUNXI_IOMEM(SUNXI_GPIO_BASE)

// PORT configuration registers
#define Pn_CFG0_REG(n)		(PIO_BASE + (n*0x24) + (0x00))    // 32bits[0..31] see data sheet
#define Pn_CFG1_REG(n)		(PIO_BASE + (n*0x24) + (0x04))
#define Pn_CFG2_REG(n)		(PIO_BASE + (n*0x24) + (0x08))
#define Pn_CFG3_REG(n)		(PIO_BASE + (n*0x24) + (0x0c))
#define Pn_CFG_REG(n, idx)	(PIO_BASE + (n*0x24) + (idx*0x04))

#define Pn_DAT(n)			(PIO_BASE + 0x0010 + (n*0x24) )

#define Pn_DRV0(n)			(PIO_BASE + 0x0014 + (n*0x24) )
#define Pn_DRV1(n)			(PIO_BASE + 0x0018 + (n*0x24) )
#define Pn_DRV(n, idx)		(PIO_BASE + (n*0x24) + 0x0014 + (idx*0x04))

#define Pn_PULL0(n)			(PIO_BASE + 0x001C + (n*0x24) )
#define Pn_PULL1(n)			(PIO_BASE + 0x0020 + (n*0x24) )
#define Pn_PULL(n, idx)		(PIO_BASE + (n*0x24) + 0x001C + (idx*0x04))

#define GPIO_BANK_A 0
#define GPIO_BANK_B 1        
#define GPIO_BANK_C 2
#define GPIO_BANK_D 3
#define GPIO_BANK_E 4
#define GPIO_BANK_F 5
#define GPIO_BANK_G 6
#define GPIO_BANK_H 7
#define GPIO_BANK_I 8

#define PIO_INT_CFG(n)		(PIO_BASE + 0x200 + (n*0x4) )

#define PIO_INT_CTL	      (PIO_BASE + 0x210)
#define PIO_INT_STA	      (PIO_BASE + 0x214)
#define PIO_INT_DEB	      (PIO_BASE + 0x218)

#define readl2(addr)        (*((volatile unsigned long  *)(addr)))
#define writel2(v, addr)    (*((volatile unsigned long  *)(addr)) = (unsigned long)(v))

#define GPIO_BANK(pin)		((pin) >> 5)
#define GPIO_NUM(pin) ((pin) & 0x1F)

static int gpiodebug = 0;

static int get_eint_no(int gpiopin)
{
	int bank = GPIO_BANK(gpiopin);
	int index = GPIO_NUM(gpiopin); 

	// PH0 .. PH21, PI10 .. PI19
	if(bank == GPIO_BANK_H)
	{
		if(index <= 21) 
			return index;
	}
	else if(bank == GPIO_BANK_I)
	{
		if((index >= 10)&&(index <= 19))
			return 22 + index;
	}
	return -1;
}

static void gpio_set_hsoc(int isactive) {
	unsigned int reg;
	unsigned int mask = 0x00000001;
	reg = readl2(PIO_INT_DEB);
	if (gpiodebug > 2){
		printk("GPIO: Old PH_EINT_DEB = 0x%.8x\n", reg);
	}
	if(isactive){
		reg |= mask;
	}
	else{
		mask = ~mask;
		reg &= mask;
	}
	writel2(reg, PIO_INT_DEB);
	reg = readl2(PIO_INT_DEB);
	if(gpiodebug > 2){
		printk("GPIO: New PH_EINT_DEB = 0x%.8x\n", reg);
	}
}


int gpio_set_pull(u32 gpiopin, u32 value)
{
	void __iomem *cfg_mem;
	u32 reg;
	u32 bank = GPIO_BANK(gpiopin);
	u32 index = GPIO_NUM(gpiopin);
	u32 cfg = (index * 2) / 32;
	u32 shift = (index * 2) - cfg * 32;
	u32 mask = ~(0x00000003 << shift);
	u32 val = (value & 0x03) << shift;

	if(cfg > 2){
		printk("dth11: GPIO: Unknown PULL register %d\n", cfg);
		return -EINVAL;
	}

	cfg_mem = IOMEM(Pn_PULL(bank, cfg));

	reg = readl2(cfg_mem);
	if (gpiodebug > 2){
		printk("GPIO: Old P%d_PULL%d = 0x%.8x\n", bank, cfg, reg);
	}
	reg &= mask;
	reg |= val;
	writel2(reg, cfg_mem);
	if (gpiodebug > 2){
		reg = readl2(cfg_mem);
		printk("GPIO: New P%d_PULL%d = 0x%.8x\n", bank, cfg, reg);
	}
	return 0;
}

int gpio_set_pin_mode(u32 gpiopin, u32 value)
{
	void __iomem *cfg_mem;
	u32 reg;
	u32 bank = GPIO_BANK(gpiopin);
	u32 index = GPIO_NUM(gpiopin);
	u32 cfg = (index * 4) / 32;
	u32 shift = (index * 4) - cfg * 32;
	u32 mask = ~(0x00000007 << shift);
	u32 val = (value & 0x07) << shift;

	if(cfg > 3){
		printk("dth11: GPIO: Unknown CFG register %d\n", cfg);
		return -EINVAL;
	}

	if(gpiodebug > 1){
		printk("GPIO: bank=%d, index=%d, cfg=%d, shift=%d, mask=0x%.8x\n", bank, index, cfg, shift, mask);
	}	

	cfg_mem = IOMEM(Pn_CFG_REG(bank, cfg));

	reg = readl2(cfg_mem);
	if (gpiodebug > 2){
		printk("GPIO: Old P%d_CFG%d = 0x%.8x\n", bank, cfg, reg);
	}
	reg &= mask;
	reg |= val;
	writel2(reg, cfg_mem);
	if (gpiodebug > 2){
		reg = readl2(cfg_mem);
		printk("GPIO: New P%d_CFG%d = 0x%.8x\n", bank, cfg, reg);
	}
	return 0;
}

int gpio_set_isr(u32 gpiopin, u32 value)
{
	u32 reg, mask;
	int index = get_eint_no(gpiopin);
	if(index >= 0){
		// enable interrupt on the pin
		if(gpiodebug > 2){
			printk("GPIO: Interrupt: %d\n", index);
		}
		reg = readl2(PIO_INT_CTL);
		if(value){
			mask = 1 << index;
			reg |= mask;
		}
		else{
			mask = ~(1 << index);
			reg &= mask;
		}
		writel2(reg, PIO_INT_CTL);
		if(gpiodebug > 2){
			printk("GPIO: PIO_INT_CTL    = 0x%.8lx\n", readl2(PIO_INT_CTL));
			printk("GPIO: PIO_INT_STA    = 0x%.8lx\n", readl2(PIO_INT_STA));
		}
	}
	else{
		printk("dth11: GPIO: Pin does not support interrupts\n");
		return -EINVAL;
	}
	return 0;
}

#endif /* _DHT11_INTERNAL_H_ */